
"""
htmlw -- Experimental HTML widget library.
"""

__version__ = '0.0.1'

# TODO: Different units for field, which override the global self.unit
#       field?  Pixels makes sense, but how should the units be set up?
# TODO: Expire page instantly.

import sys
import cgi
import string
import random
import urlparse
import webbrowser
from BaseHTTPServer import HTTPServer
from CGIHTTPServer import CGIHTTPRequestHandler

WINDOW_ID = '~_window_~'
POSITION_RELATIVE = 'relative'
POSITION_ABSOLUTE = 'absolute'
ALLOWED_POSITIONS = [POSITION_RELATIVE, POSITION_ABSOLUTE]

COORD_AUTO = None

ALLOWED_UNITS = ['em', 'ex', 'px', 'pt', 'pc', 'cm', 'mm', 'in', '%']
ALLOWED_STYLES = ['none', 'hidden', 'dotted', 'dashed', 'solid', 'double',
                  'groove', 'inset', 'outset']

ALLOWED_TILES = [True, False, 'x', 'y']

def contains_only(s, chars):
  """
  True iff s is str instance which contains only the given characters.
  """
  return isinstance(s, str) and s.translate(string.maketrans('',''), chars) == ''


def html_escape(s):
  """
  Escapes <, >, double quote, and ampersand in str s.
  """
  return s.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;')


def color_to_html(t):
  """
  Convert float color 3-tuple or 4-tuple to CSS color str (components 0.0..1.0).

    (1.0, 1.0, 1.0)      => "#ffffff"
    (1.0, 1.0, 1.0, 0.5) => 'rgba(255,255,255,0.5); filter:alpha(opacity=50)'

  Components are clamped to [0.0, 1.0].
  """
  try:
    if len(t) == 3:
      (r, g, b) = t
    elif len(t) == 4:
      (r, g, b, a) = t
  except (TypeError, AttributeError):
    raise ValueError('not a tuple of length 3 or 4: %s' % repr(t))
  if r < 0.0:
    r = 0.0
  elif r > 1.0:
    r = 1.0
  if g < 0.0:
    g = 0.0
  elif g > 1.0:
    g = 1.0
  if b < 0.0:
    b = 0.0
  elif b > 1.0:
    b = 1.0
  if len(t) == 3:
    return '#%02x%02x%02x' % (r*255,g*255,b*255)
  else:
    if a < 0.0:
      a = 0.0
    elif a > 1.0:
      a = 1.0
    return 'rgba(%d,%d,%d,%.02f); filter:alpha(opacity=%d)' % (
           r*255, g*255, b*255, a, a*100)


def is_str(x):
  """
  True iff x is a string, unicode, or buffer instance.
  """
  return isinstance(x, (str, unicode, buffer))


def parse_color(x):
  """
  Parse color object into 3-tuple or 4-tuple of floats (components 0.0...1.0).

  By "color object" we mean:

     (r, g, b)    => (r, g, b)
     '#ff00cc'    => (r, g, b)
     (r, g, b, a) => (r, g, b, a)

  Elements of the returned tuple are floats, but are not clamped to [0.0, 1.0].
  Raises ValueError if the color cannot be parsed.
  """
  if is_str(x):
    # String-like
    x = str(x).lstrip('#')
    # CSS3 colors names are not supported.
    if len(x) == 3:
      return (int(x[0]*2,16)/255.0, int(x[1]*2,16)/255.0, int(x[2]*2,16)/255.0)
    elif len(x) == 6:
      return (int(x[0:2],16)/255.0, int(x[2:4],16)/255.0, int(x[4:6],16)/255.0)
    else:
      raise ValueError('invalid color hex code: %s' % repr(x))
  elif hasattr(x, '__iter__'):
    # Tuple-like
    x = tuple(x)
    try:
      if len(x) == 3:
        return (float(x[0]), float(x[1]), float(x[2]))
      elif len(x) == 4:
        return (float(x[0]), float(x[1]), float(x[2]), float(x[3]))
    except TypeError:
      raise ValueError('invalid color: %s' % repr(x))
  raise ValueError('3-tuple or 4-tuple expected for color')


def parse_style(x):
  """
  Parse border style str x.
  """
  if hasattr(x, 'lower') and x.lower() in ALLOWED_STYLES:
    return ALLOWED_STYLES[ALLOWED_STYLES.index(x.lower())]
  raise ValueError('invalid style %s' % repr(x))


def parse_side_scalars(x):
  """
  Parse single scalar float, 4-tuple of floats, or None.
  """
  if x is None:
    return x
  try:
    return float(x)
  except (ValueError, TypeError):
    if hasattr(x, '__len__') and not is_str(x) and len(x) == 4:
      try:
        return tuple([float(y) for y in x])
      except (ValueError, TypeError):
        pass
  raise ValueError('not a float or 4 floats: %s' % repr(x))


def parse_side_styles(x):
  """
  Parse single style string, 4-tuple of style strings, or None.
  """
  if x is None:
    return x
  try:
    return parse_style(x)
  except ValueError:
    if hasattr(x, '__len__') and not is_str(x) and len(x) == 4:
      return tuple([parse_style(y) for y in x])
  raise ValueError('not a style or 4 styles: %s' % repr(x))


def parse_side_colors(x):
  """
  Parse single color (via parse_color()), 4-tuple of colors, or None.
  """
  if x is None:
    return x
  try:
    return parse_color(x)
  except ValueError:
    if hasattr(x, '__len__') and not is_str(x) and len(x) == 4:
      return tuple([parse_color(y) for y in x])
  raise ValueError('not a color or 4 colors: %s' % repr(x))


class Container(object):
  """
  Generic container, allows a parent to contain a set of children.
  """
  def __init__(self):
    self._children = set()

  def __iter__(self):
    return iter(list(self._children))

  def add(self, x):
    if hasattr(x, 'parent') and x.parent is not None:
      x.parent.remove(x)
    self._children.add(x)
    x.parent = self

  def remove(self, x):
    if not hasattr(x, 'parent') or x.parent is not self:
      raise ValueError('tried to remove child control %s with incorrect parent' % repr(x))
    self._children.remove(x)
    x.parent = None


class Contained(object):
  """
  Generic child, added to a Container instance.
  """
  def remove(self):
    if self.parent is not None:
      self.parent.remove(self)
    self.parent = None

  def __init__(self, parent):
    if parent is not None:
      parent.add(self)
    self.parent = parent


class default_property(property):
  """
  Property with specified "internal name" and default value for that name.

    default_property(internal_name, default, fset) => instance.

  If doc is not provided, then it is determined from fset's docstring.
  The property's fget is a lambda function which looks up internal_name.
  The property's fdel is not used.  This class must be used with
  property_metaclass, so that default values are bound to classes.
  """
  def __init__(self, internal_name, default, fset):
    self.internal_name = internal_name
    self.default = default
    exec 'def fget(x):\n  return x.' + internal_name
    property.__init__(self, fget=fget, fset=fset, doc=fset.__doc__)
    self.__doc__ = fset.__doc__


class property_metaclass(type):
  """
  Binds defaults of default_property().

  Also, prevents subclasses from overwriting properties with class vars.
  Works with any subclass of class default_property.
  """
  def __init__(cls, name, bases, d):
    type.__init__(cls, name, bases, d)
    # Bind defaults.
    for (k, v) in d.iteritems():
      if isinstance(v, default_property):
        setattr(cls, v.internal_name, v.default)

    # Prevent subclasses from overwriting properties.
    for k in d:
      for base in bases:
        try:
          base_prop = getattr(base, k)
        except AttributeError:
          continue
        if isinstance(base_prop, default_property):
          cls_prop = getattr(cls, k, base_prop.default)
          if not isinstance(cls_prop, default_property):
            # Call the setter with a class as its first argument.
            base_prop.fset(cls, cls_prop)
            # Change class variable (e.g. 20.0) back into a property.
            setattr(cls, k, base_prop)
            break


class Control(Contained):
  """
  Abstract base class for HTML control.
  """
  __metaclass__ = property_metaclass

  def _set_position(self, x):
    """
    Position method, either 'relative' or 'absolute' (default: 'absolute').
    """
    if x.lower() in ALLOWED_POSITIONS:
      self._position = ALLOWED_POSITIONS[ALLOWED_POSITIONS.index(x.lower())]
    else:
      raise ValueError("position is not in %s" % ', '.join(ALLOWED_POSITIONS))

  position = default_property('_position', 'absolute', _set_position)

  def _set_left(self, x):
    """
    Left coord in self.unit units, or None (default: 0).

    If the left coord is None, then the right coord will be known.
    """
    if x == COORD_AUTO:
      self._left = COORD_AUTO
      if self._right == COORD_AUTO:
        self._right = 0
    else:
      self._left = int(x)
      self._right = COORD_AUTO

  left = default_property('_left', 0, _set_left)

  def _set_top(self, x):
    """
    Top coord in self.unit units, or None (default: 0).

    If the top coord is None, then the bottom coord will be known.
    """
    if x == COORD_AUTO:
      self._top = COORD_AUTO
      if self._bottom == COORD_AUTO:
        self._bottom = 0
    else:
      self._top = int(x)
      self._bottom = COORD_AUTO

  top = default_property('_top', 0, _set_top)

  def _set_right(self, x):
    """
    Right coord in self.unit units, or None (default: None).

    If the right coord is None, then the left coord will be known.
    """
    if x == COORD_AUTO:
      self._right = COORD_AUTO
      if self._left == COORD_AUTO:
        self._left = 0
    else:
      self._right = int(x)
      self._left  = COORD_AUTO

  right = default_property('_right', None, _set_right)

  def _set_bottom(self, x):
    """
    Bottom coord in self.unit units, or None (default: None).

    If the bottom coord is None, then the top coord will be known.
    """
    if x == COORD_AUTO:
      self._bottom = COORD_AUTO
      if self._top == COORD_AUTO:
        self._top = 0
    else:
      self._bottom = int(x)
      self._top = COORD_AUTO

  bottom = default_property('_bottom', None, _set_bottom)

  def _set_width(self, x):
    """
    Width of control in self.unit units, or None (default: None).

    If None, then the width will be determined automatically by the browser.
    """
    if x == COORD_AUTO:
      self._width = COORD_AUTO
    else:
      self._width = int(x)

  width = default_property('_width', None, _set_width)

  def _set_height(self, x):
    """
    Height of control in self.unit units, or None (default: None).

    If None, then the height will be determined automatically by the browser.
    """
    if x == COORD_AUTO:
      self._height = COORD_AUTO
    else:
      self._height = int(x)

  height = default_property('_height', None, _set_height)

  def _set_zindex(self, x):
    """
    Z-index, an integer (default: 0).  Controls with higher z-indices are "on top."
    """
    self._zindex = int(x)

  zindex = default_property('_zindex', 0, _set_zindex)

  def _set_unit(self, x):
    """
    Coord units for self: one of 'em', 'ex', 'px', 'pt', 'pc', 'cm', 'mm', 'in', '%'.
    """
    if x.lower() in ALLOWED_UNITS:
      self._unit = ALLOWED_UNITS[ALLOWED_UNITS.index(x.lower())]
    else:
      raise ValueError('unit is not in %s' % ', '.join(ALLOWED_UNITS))

  unit = default_property('_unit', 'px', _set_unit)

  def _set_borderwidth(self, x):
    """
    Border width(s) in self.unit units, or None for default border.

    Either a scalar (uniform border) or a 4-tuple (top, right, bottom, left).
    """
    self._borderwidth = parse_side_scalars(x)

  borderwidth = default_property('_borderwidth', None, _set_borderwidth)

  def _set_bordercolor(self, x):
    """
    Border color(s), or None for default color.

    Pass an RGB tuple (0.5, 0.5, 0.5) for medium gray, or an RGBA tuple
    (1,1,1,0.5) for white at 50% opacity.  HTML hex codes are converted
    to RGB tuples.  Pass in a 4-tuple (top, right, bottom, left) of RGB or
    RGBA tuples, or hex codes to specify colors on each side.
    """
    self._bordercolor = parse_side_colors(x)

  bordercolor = default_property('_bordercolor', None, _set_bordercolor)

  def _set_borderstyle(self, x):
    """
    Border style(s), or None for default style.

    Valid styles are 'none', 'hidden', 'dotted', 'dashed', 'solid',
    'double', 'groove', 'ridge', 'inset', 'outset'.  Note that the Python
    None value causes the browser to use its default style, whereas
    'none' causes the border to disappear.

    Either a single string (uniform border styles) or a 4-tuple of
    style strings (top, right, bottom, left).
    """
    self._borderstyle = parse_side_styles(x)

  borderstyle = default_property('_borderstyle', None, _set_borderstyle)

  def _set_padding(self, x):
    """
    Padding width(s) (default: 0).

    Either a scalar (uniform padding) or a 4-tuple (top, right, bottom, left).
    """
    if x is None:
      x = 0.0
    self._padding = parse_side_scalars(x)

  padding = default_property('_padding', 0.0, _set_padding)

  def _set_forecolor(self, x):
    """
    Foreground color, or None to use default (default: None).

    Pass an RGB tuple (0.5, 0.5, 0.5) for medium gray, or an RGBA tuple
    (1,1,1,0.5) for white at 50% opacity.  HTML hex codes are converted
    to RGB tuples.
    """
    if x is None:
      self._forecolor = None
    else:
      self._forecolor = parse_color(x)

  forecolor = default_property('_forecolor', None, _set_forecolor)

  def _set_backcolor(self, x):
    """
    Background color, or None to use default (default: None).

    Pass an RGB tuple (0.5, 0.5, 0.5) for medium gray, or an RGBA tuple
    (1,1,1,0.5) for white at 50% opacity.  HTML hex codes are converted
    to RGB tuples.
    """
    if x is None:
      self._backcolor = None
    else:
      self._backcolor = parse_color(x)

  backcolor = default_property('_backcolor', None, _set_backcolor)

  def _set_image(self, x):
    """
    Background image URL, or '' to use default (default: '').
    """
    self._image = str(x)

  image = default_property('_image', '', _set_image)

  def _set_imagetile(self, x):
    """
    Background image tiling (default: True).

    Use False to disable tiling, True to tile in X and Y directions,
    'x' to tile in X direction only, 'y' to tile in Y direction only.
    """
    y = x
    if hasattr(y, 'lower'):
      y = y.lower()
    if y in ALLOWED_TILES:
      self._imagetile = ALLOWED_TILES[ALLOWED_TILES.index(y)]
    else:
      raise ValueError('%s is not in %r' % (repr(x), ALLOWED_TILES))

  imagetile = default_property('_imagetile', True, _set_imagetile)

  def _set_imagescroll(self, x):
    """
    Background image scrolling (default: True).

    Use False to disable the image from scrolling in the browser.
    """
    if x in [True, False]:
      self._imagescroll = bool(x)
    else:
      raise ValueError('%s is not a bool' % repr(x))

  imagescroll = default_property('_imagescroll', True, _set_imagescroll)

  def _set_imageleft(self, x):
    """
    Background image left coord (default: 0).

    Note that there is no right coord for images, due to the CSS spec.
    """
    self._imageleft = int(x)

  imageleft = default_property('_imageleft', 0, _set_imageleft)

  def _set_imagetop(self, x):
    """
    Background image top coord (default: 0).

    Note that there is no bottom coord for images, due to the CSS spec.
    """
    self._imagetop = int(x)

  imagetop = default_property('_imagetop', 0, _set_imagetop)

  def _set_name(self, x):
    """
    HTML name for control: alphanumeric and underscores in ASCII only.
    """
    e = ValueError('invalid name: %s' % repr(x))
    try:
      x = str(x.encode('ascii'))
    except:
      raise e
    if len(x) == 0:
      raise ValueError('empty name')
    if contains_only(x, string.ascii_letters + string.digits + '_'):
      self._name = x
    else:
      raise e

  name = default_property('_name', '', _set_name)

  def get_style(self):
    """
    Get CSS style attribute.
    """
    def add_coord_if_not_none(k):
      if getattr(self, k) is not None:
        ans.append('%s: %s%s' % (k, getattr(self, k), self.unit))
    ans = []
    ans.append('position: %s' % self.position)
    for k in 'left top right bottom width height'.split():
      add_coord_if_not_none(k)
    ans.append('z-index: %d' % self.zindex)
    if self.borderwidth is not None:
      if isinstance(self.borderwidth, float):
        ans.append('border-width: %s%s' % (self.borderwidth, self.unit))
      else:
        ans.append('border-width: %s' % ' '.join(['%d%s' % (x, self.unit)
                                                  for x in self.borderwidth]))
    if self.bordercolor is not None:
      if isinstance(self.bordercolor[0], float):
        ans.append('border-color: %s' % color_to_html(self.bordercolor))
      else:
        ans.append('border-color: %s' % ' '.join([color_to_html(x)
                                                  for x in self.bordercolor]))
    if self.borderstyle is not None:
      if isinstance(self.borderstyle, str):
        ans.append('border-style: %s' % self.borderstyle)
      else:
        ans.append('border-style: %s %s %s %s' % self.borderstyle)

    if self.padding is not None:
#      print 'Padding is:', self.padding
      if isinstance(self.padding, float):
        ans.append('padding: %s%s' % (self.padding, self.unit))
      else:
        ans.append('padding: %s' % ' '.join(['%d%s' % (x, self.unit)
                                             for x in self.padding]))

    if self.forecolor is not None:
      ans.append('color: %s' % color_to_html(self.forecolor))

    if self.backcolor is not None:
      ans.append('background-color: %s' % color_to_html(self.backcolor))

    if self.image != '':
      ans.append('background-image: url(%s)' %
                 self.image.replace('"', '%22').replace("'", '%27'))

    if self.imagetile != True:
      repeat_map = {True: 'repeat', False: 'no-repeat', 'x': 'repeat-x',
                    'y': 'repeat-y'}
      ans.append('background-repeat: %s' % repeat_map[self.imagetile])

    if self.imagescroll != True:
      ans.append('background-attachment: fixed')

    if (self.imageleft, self.imagetop) != (0, 0):
      ans.append('background-position: %d%s %d%s' %
                 (self.imageleft, self.unit, self.imagetop, self.unit))

    return '; '.join(ans)

  def get_html(self):
    """
    Get HTML string for control (not implemented in base class).

    Controls should never be added nor removed from the Window instance
    containing self during get_html().
    """
    return ''

  def on_submit(self, d):
    """
    Change properties in response to form submission.

    Here d is a cgi.FieldStorage instance.
    """

  def on_postsubmit(self, d):
    """
    Call user event handlers, in a separate stage after on_submit().

    Here d is a cgi.FieldStorage instance.
    """

  def __init__(self, parent, **kw):
    Contained.__init__(self, parent)
    for (k, v) in kw.items():
      setattr(self, k, v)
    if self.name == '':
      self.name = '%s_%08x' % (self.__class__.__name__, id(self))


class ControlWithText(Control):
  """
  Control with string text property.
  """
  def _set_text(self, x):
    """
    Control's visible text.
    """
    self._text = str(x)

  text = default_property('_text', '', _set_text)

  def __init__(self, *args, **kw):
    Control.__init__(self, *args, **kw)
    if self.text == '':
      self.text = self.__class__.__name__


class ContainerControl(Control, Container):
  """
  Control which can contain sub-elements.
  """
  def __init__(self, *args, **kw):
    Control.__init__(self, *args, **kw)
    Container.__init__(self)

  def get_html(self):
    ans = []
    for x in self:
      ans += [x.get_html()]
    return '\n'.join(ans)

  def on_submit(self, d):
    for x in self:
      x.on_submit(d)

  def on_postsubmit(self, d):
    for x in self:
      x.on_postsubmit(d)


class Button(ControlWithText):
  """
  Button HTML control.
  """
  def get_html(self):
    return ('<input type="submit" name="%s" style="%s" value="%s">' %
            (self.name, self.get_style(), html_escape(self.text)))

  def on_postsubmit(self, d):
    if self.name in d:
      self.on_click()

  def on_click(self):
    pass


class Textbox(ControlWithText):
  def _set_multiline(self, x):
    """
    Multiline, True or False (default: False).

    Setting multiline to True will disable password input mode.
    """
    self._multiline = bool(x)
    if self._multiline:
      self._password = False

  multiline = default_property('_multiline', False, _set_multiline)

  def _set_password(self, x):
    """
    Password input box, True or False (default: False).

    Setting password to True will disable multiline mode.
    """
    self._password = bool(x)
    if self._password:
      self._multiline = False

  password = default_property('_password', False, _set_password)

  def _set_readonly(self, x):
    """
    Is box read-only?  True or False (default: False).
    """
    self._readonly = bool(x)

  readonly = default_property('_readonly', False, _set_readonly)

  def on_submit(self, d):
    if self.name in d:
      self.text = d[self.name].value

  def get_html(self):
    attrs = ''
    if self.readonly:
      attrs = 'readonly'
    if self.multiline:
      return ('<textarea name="%s" style="%s" %s>%s</textarea>' %
              (self.name, self.get_style(), attrs, html_escape(self.text)))
    else:
      password_map = {False: 'text', True: 'password'}
      return ('<input type="%s" name="%s" style="%s" %s value="%s">' %
              (password_map[self.password], self.name, self.get_style(), attrs, html_escape(self.text)))


class Checkbox(ControlWithText):
  def _set_checked(self, x):
    """
    Is box checked?  True or False (default: False).
    """
    self._checked = bool(x)

  checked = default_property('_checked', False, _set_checked)

  def get_html(self):
    checked = ''
    if self.checked:
      checked = 'checked'
    return ('<div style="%s"><input type="checkbox" name="%s" %s>%s</div>' %
            (self.get_style(), self.name, checked, self.text))

  def on_submit(self, d):
    if self.name in d:
      self.checked = True
    else:
      self.checked = False


class Radio(ControlWithText):
  def _set_checked(self, x):
    """
    Is radio checked?  True or False (default: False).
    """
    self._checked = bool(x)

  checked = default_property('_checked', False, _set_checked)

  def _set_group(self, x):
    """
    Group of buttons that radio belongs to.

    ASCII string containing alphanumeric and underscores only.
    By default, each radio's group is set to its name.  Thus by
    default, each radio button belongs to its own individual group.
    """
    self._group = str(x)

  group = default_property('_group', '', _set_group)

  def get_html(self):
    checked = ''
    if self.checked:
      checked = 'checked'
    return ('<div style="%s"><input type="radio" name="%s" value="%s" %s>%s</div>' %
            (self.get_style(), self.group, self.name, checked, self.text))

  def on_submit(self, d):
    if self.group in d:
      self.checked = d[self.group].value == self.name

  def __init__(self, parent, **kw):
    ControlWithText.__init__(self, parent, **kw)
    if self.group == '':
      self.group = self.name


class Frame(ContainerControl):
  """
  Frame (i.e. a <div>).
  """
  def _set_text(self, x):
    """
    Control's visible text.
    """
    self._text = str(x)

  text = default_property('_text', '', _set_text)

  def get_html(self):
    return ('<div style="%s" name="%s">%s%s</div>' %
            (self.get_style(), self.name, self.text,
             ContainerControl.get_html(self)))


def open_browser(window):
  """
  Open platform-specific browser.

  Here window is a Window instance.
  """
  if sys.platform == 'win32' and False:
    open_ie_browser(window)
  else:
    open_generic_browser(window)


def open_ie_browser(window):
  """
  Open non-resizable IE browser on given URL (does not work).
  """
  """
  import threading
  def main():
    import wx
    from wx.lib.iewin import IEHtmlWindow

    class IEFrame(wx.Frame):
      def __init__(self, parent):
        wx.Frame.__init__(self, parent, -1, window.title, size=(window.width, window.height),
                                style=wx.MINIMIZE_BOX | wx.SYSTEM_MENU | wx.CAPTION | wx.CLOSE_BOX | wx.CLIP_CHILDREN)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.ie = IEHtmlWindow(self, -1)
        self.ie.LoadUrl(window.url)
        sizer.Add(self.ie, 1, wx.EXPAND)
        self.SetSizer(sizer)
        self.Show(True)

    app = wx.PySimpleApp(0)
    frame = IEFrame(None)
    app.MainLoop()

  global T
  T = threading.Thread(target=main, args=())
  T.start()
  """

def open_generic_browser(window):
  """
  Open generic browser via webbrowser module.
  """
  import webbrowser
  webbrowser.open(window.url, 1)


class ServerHandler(CGIHTTPRequestHandler):
  def handle_window(self, window_name):
#    self.send_header('Expires', '0')
#    self.send_header('Pragma', 'no-cache')
#    self.send_header('Cache-Control', 'no-cache')

    query = ''
    if '?' in self.path:
      query = self.path[self.path.index('?')+1:]
    env = {}
    env['REQUEST_METHOD'] = self.command
    env['QUERY_STRING']   = query

    # From CGIHTTPServer.py
    if self.headers.typeheader is None:
      env['CONTENT_TYPE'] = self.headers.type
    else:
      env['CONTENT_TYPE'] = self.headers.typeheader
    n = self.headers.getheader('content-length')
    if n:
      env['CONTENT_LENGTH'] = n

    d = cgi.FieldStorage(self.rfile, environ=env)

    windows = [x for x in self.server if x.name == window_name]
    if len(windows) == 0:
      raise ValueError('no window with name %s' % window_name)
    elif len(windows) >= 2:
      raise ValueError('%d windows with name %s' % (len(windows), window_name))
    else:
      windows[0].on_submit(d)
      windows[0].on_postsubmit(d)
      self.wfile.write(windows[0].get_html())

  def get_window_name(self):
    window_path = urlparse.urlparse(self.path)[2]
    if window_path.startswith('/'):
      window_path = window_path[1:]
    if not window_path.startswith(WINDOW_ID):
      return None
    return window_path[len(WINDOW_ID):]

  def is_cgi(self):
    w = self.get_window_name()
    return (w is not None) or CGIHTTPRequestHandler.is_cgi(self)

  def run_cgi(self):
    w = self.get_window_name()
    if w is not None:
      self.handle_window(w)
    else:
      CGIHTTPRequestHandler.run_cgi(self)


class Server(HTTPServer, Container):
  """
  Simple HTTP server which handles requests from Window instances.
  """
  def __init__(self, port=None, HandlerClass=ServerHandler):
    Container.__init__(self)
    if port is None:
      port = random.randrange(8000, 16000)
    self.port = port
    HTTPServer.__init__(self, ('127.0.0.1', port), ServerHandler)
    self.visible = False

  def get_window_url(self, window):
    return 'http://127.0.0.1:%d/%s%s' % (self.port, WINDOW_ID, window.name)

  def show(self):
    if self.visible == False:
      self.visible = True
      for x in self:
        open_browser(x)

  def handle_request(self):
    self.show()
    HTTPServer.handle_request(self)


def get_default_server():
  """
  Get default server instance, used when Window() initialized with server=None.
  """
  global _default_server
  try:
    return _default_server
  except NameError:
    _default_server = Server()
    return _default_server


def set_default_server(s):
  """
  Set default server instance, used when Window() initialized with server=None.
  """
  global _default_server
  _default_server = s


class Window(Container, Contained):
  """
  Browser window which renders the given HTML controls.
  """
  name       = ''
  method     = 'GET'
  title      = 'Window title'
  body_attrs = ''
  head       = ''

  def get_html(self):
#    print 'Getting HTML for window...'
    ans = ['<html><head><title>%s</title>%s</head><body %s>' %
           (self.title, self.head, self.body_attrs)]
    ans += ['<form action="%s" method="%s"><input type="hidden" name="~loaded" value="1">' % (self.url, self.method)]
    for x in self:
      ans.append(x.get_html())
    ans += ['</form></body></html>']
    return '\n'.join(ans)

  url = property(fget=lambda self: self.parent.get_window_url(self))

  def __init__(self, parent=None, **kw):
    if parent is None:
      parent = get_default_server()
    Container.__init__(self)
    Contained.__init__(self, parent)
    for (k, v) in kw.items():
      setattr(self, k, v)
    if self.name == '':
      self.name       = 'window_%08x' % id(self)

  def on_submit(self, d):
#    print 'Submitted window', self.name
    for x in self:
      x.on_submit(d)

  def on_postsubmit(self, d):
#    print 'Calling event handlers for window', self.name
    for x in self:
      x.on_postsubmit(d)
