
"""
Minesweeper, with suggested moves by propositional logic.
"""

__version__ = '0.8.0'

import copy
import random
import os
print os.getcwd()

from entail import AssumptionSet
from htmlw import *

class Minesweeper:
  """
  Minesweeper game instance.
  """
  def __init__(self, w=9, h=9, mines=10):
    self.w = int(w)
    self.h = int(h)
    self.mines = int(mines)
    self.lost = False
    self.won = False

    if self.w < 1 or self.h < 1:
      raise ValueError('invalid w, h: %r' % ((self.w, self.h),))
    if self.mines < 1 or self.mines > self.w * self.h:
      raise ValueError('invalid mines value: %d' % self.mines)

    self.mine = [[False for x in range(self.w)]
                        for y in range(self.h)]
    self.chosen = copy.deepcopy(self.mine)

    unoccupied = [(x, y) for x in range(self.w)
                         for y in range(self.h)]
    for (x, y) in random.sample(unoccupied, self.mines):
      self.mine[y][x] = True

  def in_bounds(self, x, y):
    """
    True iff (x, y) is valid position on board.
    """
    return 0 <= x < self.w and 0 <= y < self.h

  def neighbors(self, x, y):
    """
    Iterator of (x', y') tuples with inf-norm distance <= 1.
    """
    for dy in [-1, 0, 1]:
      for dx in [-1, 0, 1]:
        if (dx, dy) != (0, 0):
          if self.in_bounds(x+dx, y+dy):
            yield (x+dx, y+dy)

  def neighboring_mines(self, x, y):
    """
    Get number of neighboring mines for (x, y).
    """
    ans = 0
    for (xp, yp) in self.neighbors(x, y):
      ans += self.mine[yp][xp]
    return ans

  def get_text(self, x, y):
    """
    Get visible text at given square: '0' to '8', '?', or '*'.
    """
    if not self.in_bounds(x, y):
      raise ValueError('invalid (x, y): %r' % ((x, y),))
    if self.chosen[y][x] or (self.won or self.lost):
      if self.mine[y][x]:
        return '*'
      else:
        return str(self.neighboring_mines(x, y))
    return '?'

  def choose(self, x, y):
    """
    Make move at given square.

    Raises ValueError if move cannot be made.  Check for game-over
    after calling choose() with self.lost and self.won.
    """
    if not self.in_bounds(x, y):
      raise ValueError('invalid (x, y): %r' % ((x, y),))
    if self.chosen[y][x]:
      raise ValueError('move already made: %r' % ((x, y),))
    if self.lost or self.won:
      raise ValueError('cannot make move: game is over')
    if self.mine[y][x]:
      self.lost = True

    def enable_square(xc, yc):
      self.chosen[yc][xc] = True
      if self.neighboring_mines(xc, yc) == 0:
        for (xp, yp) in self.neighbors(xc, yc):
          if not self.chosen[yp][xp] and not self.mine[yp][xp]:
            enable_square(xp, yp)

    enable_square(x, y)
    if not self.lost and sum([sum(row) for row in self.chosen]) == self.w*self.h-self.mines:
      self.won = True

  def suggest(self):
    """
    Return suggested (x, y) move, or raise ValueError if no suggestion.
    """
    assert self.w < 1000 and self.h < 1000
    def mine_str(x, y):
      return 'm%03d_%03d' % (y, x)

    # Add all knowledge of mines (what the user sees).
    s = AssumptionSet()
    for y in range(self.h):
      for x in range(self.w):
        v = self.get_text(x, y)
        if v == '*':
          s.assume(mine_str(x, y))
        elif v == '?':
          pass
        else:
          s.assume('not ' + mine_str(x, y))
          # If v == '2' and x=0, y=0, then add an assumption such as
          # "m_001_000 + m_000_001 + m_001_001 == 2".
          s.assume('+'.join([mine_str(*p) for p in
                          self.neighbors(x, y)]) + '==' + v)

    # Try to deduce a good move from knowledge of mines.
    for y in range(self.h):
      for x in range(self.w):
        if (self.get_text(x, y) == '?' and
            s.implies('not ' + mine_str(x, y))):
          return (x, y)
    raise ValueError('no suggested move')


# -------------------------------------------------------------------
# HTML interface to the Minesweeper class.
# -------------------------------------------------------------------

class MinesweeperButton(Button):
  def on_click(self):
    self.parent.parent.on_click(self.x, self.y)


class SuggestButton(Button):
  def on_click(self):
    self.parent.parent.on_suggest_click()


class MinesweeperWindow(Minesweeper, Window):
  width  = 408
  height = 570
  title  = 'Minesweeper'

  def __init__(self, **kw):
    Minesweeper.__init__(self, **kw)
    Window.__init__(self)
    w = 30
    h = 30
    padding = 5
    outerw = w+padding*2
    outerh = h+padding*2
    self.default_status = 'Click a mine to play'
    self.suggest_xy = (-1, -1)
    self.title_frame = Frame(self, left=25, top=25/2, width=300, height=75, image='mine.png')
    self.frame = Frame(self, left=25, top=self.title_frame.top+self.title_frame.height+15)
    self.button = [[MinesweeperButton(self.frame,
                           left=outerw*x, top=outerh*y,
                           width=w, height=h, padding=padding,
                           x=x, y=y, borderwidth=1)
                    for x in range(self.w)]
                    for y in range(self.h)]
    self.label = Textbox(self.frame, text=self.default_status,
                         left=0, top=outerh*self.h+10, readonly=True,
                         width=170)
    self.suggest_btn = SuggestButton(self.frame, text='Suggest Move',
                                     left=170+10, top=outerh*self.h+10)
    self.update_buttons()

  def update_buttons(self):
    colormap = {1: '#BE9EB8', 2: '#6285B4', 3: '#659C5C',
                4: '#DCD492', 5: '#D7AE85', 6: '#B96F6F'}
    colormap[7] = colormap[8] = colormap[6]

    bright = '#ECE9D8'
    light = (0.7,)*3
    dark = (0.5,)*3
    done = '#98BCB8'
    for y in range(self.h):
      for x in range(self.w):
        b = self.button[y][x]
        b.borderstyle = 'solid'
        b.borderwidth = 1
        v = self.get_text(x, y)
        if v == '?':
          v = ' '
          b.backcolor = bright
          b.bordercolor = (light,dark,dark,light)
        elif v == '*':
          if self.won:
            b.backcolor = done
          else:
            b.backcolor = (1,0.2,0.2)
        else:
          b.backcolor = done
          b.bordercolor = (dark,dark,dark,dark)
          if v == '0':
            v = ''
          else:
            b.backcolor = colormap[int(v)]
        if (x, y) == self.suggest_xy:
          b.borderstyle = 'dashed'
          b.borderwidth = 2
          b.bordercolor = (1,0,0)
        b.text = v

  def on_suggest_click(self):
    if self.won or self.lost:
      self.__init__()
    else:
      try:
        self.suggest_xy = self.suggest()
        self.label.text = self.default_status
      except ValueError:
        self.label.text = 'No suggestion'
        self.suggest_xy = (-1, -1)
      self.update_buttons()    

  def on_click(self, x, y):
    self.suggest_xy = (-1, -1)
    if not (self.lost or self.won):
      try:
        self.choose(x, y)
      except ValueError:
        return
      self.update_buttons()
    if self.lost:
      self.label.text = 'You lose.'
      self.suggest_btn.text = 'New Game'
    if self.won:
      self.label.text = 'You win!'
      self.suggest_btn.text = 'New Game'

  def get_html(self):
    return Window.get_html(self)


def main():
  window = MinesweeperWindow()
  get_default_server().serve_forever()


if __name__ == '__main__':
  main()
